@@ -38,7 +38,7 @@ import os
 import re
 import shutil
 import site
-import subprocess
+import requests
 import sys
 import tarfile
 import tempfile
@@ -49,10 +49,12 @@ import uuid
 import yaml
 import yamlordereddictloader
 import zipfile
+import subprocess
+
 
 from abc import ABC, abstractmethod
-from fuzzywuzzy import fuzz
-from fuzzywuzzy import process as fuzzprocess
+from rapidfuzz import fuzz
+from rapidfuzz import process as fuzzprocess
 from mlhub.constants import (
     APP,
     APPX,
@@ -377,9 +379,9 @@ def get_url_filename(url):
 
     # Specify header to avoid a 403 from some sites.
     headers = {'User-Agent':
-               'Mozilla/5.0 (X11; Linux x86_64) ' +
-               'AppleWebKit/537.36 (KHTML, like Gecko) ' +
-               'Chrome/81.0.4044.138 Safari/537.36'}
+                   'Mozilla/5.0 (X11; Linux x86_64) ' +
+                   'AppleWebKit/537.36 (KHTML, like Gecko) ' +
+                   'Chrome/81.0.4044.138 Safari/537.36'}
     req = urllib.request.Request(url, headers=headers)
     try:
         info = urllib.request.urlopen(req).getheader("Content-Disposition")
@@ -387,7 +389,7 @@ def get_url_filename(url):
         if err.code == 404:
             print(f"\nmlhub: Missing url: {url}\n       please notify package author.")
             return None
-        
+
     if info:
         _, params = cgi.parse_header(info)
         if "filename" in params:
@@ -427,6 +429,54 @@ def download_model_pkg(url, local, pkgfile, quiet):
         raise ModelDownloadHaltException(url, error.reason.lower())
 
 
+# ----------------------------------------------------------------------
+# Check repo default branch
+# ----------------------------------------------------------------------
+
+def check_default_branch(owner, repo, repo_type):
+    """Check repository default branch."""
+
+    check_github_url = "https://api.github.com/repos/{}/{}".format(
+        owner, repo
+    )
+
+    check_gitlab_url = "https://gitlab.com/api/v4/projects/{}%2F{}".format(
+        owner, repo
+    )
+
+    check_bitbucket_url = "https://api.bitbucket.org/2.0/repositories/{}/{}".format(
+        owner, repo
+    )
+    # For github
+    if repo_type == "github":
+        try:
+            ref = requests.get(check_github_url).json()['default_branch']
+            return ref
+        except Exception as e:
+            print(f"Error: {e}\nThe project doesn't exist. ")
+            sys.exit(1)
+
+    # For gitlab
+    elif repo_type == "gitlab":
+        try:
+            ref = requests.get(check_gitlab_url).json()['default_branch']
+            return ref
+        except Exception as e:
+            print(f"Error: {e}\nThe project doesn't exist. ")
+            sys.exit(1)
+
+    # For bitbucket
+    elif repo_type == "bitbucket":
+        try:
+            ref = requests.get(check_bitbucket_url).json()['mainbranch']['name']
+            return ref
+        except Exception as e:
+            print(f"Error: {e}\nThe project doesn't exist. ")
+            sys.exit(1)
+
+
+
+
 # ----------------------------------------------------------------------
 # Folder and file manipulation
 # ----------------------------------------------------------------------
@@ -494,8 +544,8 @@ def unpack_with_promote(file, dest, valid_name=None, remove_dst=True):
         file_list = getattr(pkg_file, lister_name)()
         first_segs = [x.split(os.path.sep)[0] for x in file_list]
         if (len(file_list) == 1 and os.path.sep in file_list[0]) or (
-            len(file_list) != 1
-            and all([x == first_segs[0] for x in first_segs])
+                len(file_list) != 1
+                and all([x == first_segs[0] for x in first_segs])
         ):
             promote, top_dir = True, file_list[0].split(os.path.sep)[0]
         else:
@@ -530,7 +580,7 @@ def unpack_with_promote(file, dest, valid_name=None, remove_dst=True):
                     # existing files except they have the same name.
 
                     with opener(
-                        os.path.join(tmpdir2, "tmpball"), "w"
+                            os.path.join(tmpdir2, "tmpball"), "w"
                     ) as new_pkg_file:
                         appender = getattr(new_pkg_file, appender_name)
                         dir_path = os.path.join(tmpdir, top_dir)
@@ -542,7 +592,7 @@ def unpack_with_promote(file, dest, valid_name=None, remove_dst=True):
                                 appender(file_path, arc_path)
 
                     with opener(
-                        os.path.join(tmpdir2, "tmpball")
+                            os.path.join(tmpdir2, "tmpball")
                     ) as new_pkg_file:
                         new_pkg_file.extractall(dest)
 
@@ -610,7 +660,7 @@ def is_tar(name):
     """Check if name is a Tarball."""
 
     return (
-        name.endswith(".tar") or name.endswith(".gz") or name.endswith(".bz2")
+            name.endswith(".tar") or name.endswith(".gz") or name.endswith(".bz2")
     )
 
 
@@ -624,9 +674,9 @@ def is_description_file(name):
     """Check if name ends with DESCRIPTION.yaml or DESCRIPTION.yml"""
 
     return (
-        name.endswith(DESC_YAML)
-        or name.endswith(DESC_YML)
-        or name.endswith(MLHUB_YAML)
+            name.endswith(DESC_YAML)
+            or name.endswith(DESC_YML)
+            or name.endswith(MLHUB_YAML)
     )
 
 
@@ -699,7 +749,7 @@ def get_command_suggestion(cmd, description=None, model=""):
         msg = meta.get(
             "suggestion",
             "\nTo " + dropdot(lower_first_letter(meta["description"])) + ":"
-            "\n\n  $ {} {} {}",
+                                                                         "\n\n  $ {} {} {}",
         )
         msg = msg.format(CMD, cmd, model)
         return msg
@@ -928,7 +978,7 @@ def install_r_deps(deps, model, source="cran", yes=False):
 def install_python_deps(deps, model, source="pip", yes=False):
     logger = logging.getLogger(__name__)
 
-    env_var  = 'export _MLHUB_OPTION_YES="y"; ' if yes else ""
+    env_var = 'export _MLHUB_OPTION_YES="y"; ' if yes else ""
     env_var += 'export _MLHUB_PYTHON_EXE="{}"; '.format(sys.executable)
 
     script = os.path.join(os.path.dirname(__file__),
@@ -964,7 +1014,7 @@ def install_python_deps(deps, model, source="pip", yes=False):
             return
 
         jd = '" "'.join(deps) if isinstance(deps, list) else deps
-        command  = f'{env_var}{BASH_CMD} {script} "{pkg_dir}" '
+        command = f'{env_var}{BASH_CMD} {script} "{pkg_dir}" '
         command += f'{source} {category} "{jd}"'
     else:
         if source.startswith("pip"):
@@ -976,9 +1026,9 @@ def install_python_deps(deps, model, source="pip", yes=False):
     logger.debug(f"INSTALL command: {command}")
 
     proc = subprocess.Popen(command,
-                            shell  = True,
-                            cwd    = get_package_dir(model),
-                            stderr = subprocess.PIPE)
+                            shell=True,
+                            cwd=get_package_dir(model),
+                            stderr=subprocess.PIPE)
     output, errors = proc.communicate()
     if proc.returncode != 0:
         errors = errors.decode("utf-8")
@@ -1158,7 +1208,7 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
         # configure`.
 
         if downloadir is None and (
-            is_url(location) or RepoTypeURL.is_repo_ref(location)
+                is_url(location) or RepoTypeURL.is_repo_ref(location)
         ):
 
             # URL for non-package files
@@ -1190,7 +1240,6 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
                 )  # The name of the file to be downloaded
 
                 if filename is None:
-
                     # TODO: The file name cannot be determined from URL.
                     #       How to deal with this scenario?  Current
                     #       solution: We give it a random name.  This
@@ -1216,12 +1265,12 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
                 else:
                     if filetype == "file":
                         if (
-                            target.endswith(os.path.sep) and not is_archive
+                                target.endswith(os.path.sep) and not is_archive
                         ):  # Download into a specified folder
                             target = os.path.join(target, filename)
                     else:
                         if target.endswith(
-                            os.path.sep
+                                os.path.sep
                         ):  # Unzip repo/dir into a folder with the same name
                             target = os.path.join(target, foldername, "")
                         else:  # Unzip repo/dir into a folder with a different name
@@ -1229,7 +1278,7 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
 
                 if target.endswith(os.path.sep):  # Expand path
                     target = (
-                        os.path.relpath(target) + os.path.sep
+                            os.path.relpath(target) + os.path.sep
                     )  # Ensure folder end with '/'
                 else:
                     target = os.path.relpath(target)
@@ -1258,7 +1307,6 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
                 download_msg = "      downloading into {} ..."
 
                 if os.path.exists(archive):
-
                     # 20190327 gjw for now cache management is behind
                     # scenes and do not need to ask for each one. If
                     # already in cache then don't download. If user wants
@@ -1314,9 +1362,9 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
                     make_symlink(origin, goal)
 
         if (
-            downloadir is not None
-            and not (is_url(location) or RepoTypeURL.is_repo_ref(location))
-            or maybe_private
+                downloadir is not None
+                and not (is_url(location) or RepoTypeURL.is_repo_ref(location))
+                or maybe_private
         ):
 
             # Path for package files or private Git repo
@@ -1326,7 +1374,7 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
             try:
                 goal = os.path.join(pkg_dir, "" if target is None else target)
                 if location.endswith(
-                    "*"
+                        "*"
                 ):  # Move all files under <location> to package's root dir
                     origin = os.path.join(downloadir, location[:-2])
                     merge_folder(origin, goal)
@@ -1362,7 +1410,7 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
                         else:
                             origin = os.path.join(downloadir, location)
                         if os.path.isdir(origin) and not goal.endswith(
-                            os.path.sep
+                                os.path.sep
                         ):
                             merge_folder(origin, goal)
                         else:
@@ -1378,11 +1426,10 @@ def install_file_deps(deps, model, downloadir=None, key=None, yes=False):
 
 
 class RepoTypeURL(ABC):
-
     REPO_DOMAINS = {
-        "github": ["github.com", "githubusercontent.com",],
-        "gitlab": ["gitlab.com",],
-        "bitbucket": ["bitbucket.org",],
+        "github": ["github.com", "githubusercontent.com", ],
+        "gitlab": ["gitlab.com", ],
+        "bitbucket": ["bitbucket.org", ],
     }
 
     def __init__(self, url, prefix, ssh_host):
@@ -1412,14 +1459,14 @@ class RepoTypeURL(ABC):
         url = self.url
         prefix_len = len(self.prefix)
         if url.lower().startswith(
-            self.prefix
+                self.prefix
         ):  # Remove prefix 'prefix[@xxx]:'
             if url[prefix_len] in "@":
                 colon_index = url.index(":")
-                self.ssh_host = url[prefix_len + 1 : colon_index]
-                url = url[colon_index + 1 :].strip()
+                self.ssh_host = url[prefix_len + 1: colon_index]
+                url = url[colon_index + 1:].strip()
             elif url[prefix_len] in ":":
-                url = url[prefix_len + 1 :].strip()
+                url = url[prefix_len + 1:].strip()
         return url
 
     @staticmethod
@@ -1490,11 +1537,12 @@ class RepoTypeURL(ABC):
             return False
 
     @staticmethod
-    def interpret_repo_ref(url):
+    def interpret_repo_ref(url, repo_type):
 
         # TODO: Add regexp to validate the url.
 
-        ref = "master"  # Use master by default.
+        # ref = "master"  # Use master by default.
+
         path = None  # Path to a specific file, if not, the repo.
 
         if ":" in url:  # Get path
@@ -1502,6 +1550,8 @@ class RepoTypeURL(ABC):
 
         owner, repo = url.split("/")
 
+        ref = check_default_branch(owner, repo, repo_type)
+
         if "@" in repo:  # Branch or commit: mlhub@dev
             repo, ref = repo.split("@")
         elif "#" in repo:  # Pull request: mlhub#15
@@ -1550,6 +1600,8 @@ class GitHubURL(RepoTypeURL):
         #   return "https://api.github.com/repos/{}/{}/zipball/{}".format(
         #       self.owner, self.repo, self.ref)
 
+        self.ref = check_default_branch(self.owner, self.repo, "github")
+
         return "https://codeload.github.com/{}/{}/zip/{}".format(
             self.owner, self.repo, self.ref
         )
@@ -1557,6 +1609,8 @@ class GitHubURL(RepoTypeURL):
     def compose_content_url(self, api=False, tree=False, default="{}"):
         """Compose GitHub URL for the content of a file or a directory."""
 
+        self.ref = check_default_branch(self.owner, self.repo, "github")
+
         if api or self.ref.startswith("pull/"):
             return "https://api.github.com/repos/{}/{}/contents/{}?ref={}".format(
                 self.owner, self.repo, self.path or default, self.ref
@@ -1616,7 +1670,7 @@ class GitHubURL(RepoTypeURL):
         url = self.remove_prefix()
 
         if not is_url(
-            url
+                url
         ):  # Reference such as mlhubber/mlhub@dev:doc/MLHUB.yaml
 
             (
@@ -1624,7 +1678,7 @@ class GitHubURL(RepoTypeURL):
                 self.repo,
                 self.ref,
                 self.path,
-            ) = RepoTypeURL.interpret_repo_ref(url)
+            ) = RepoTypeURL.interpret_repo_ref(url, "github")
 
             if "#" in url:
                 self.ref = "pull/" + self.ref + "/head"
@@ -1667,6 +1721,8 @@ class GitLabURL(RepoTypeURL):
     def compose_repo_zip_url(self):
         """Compose GitLab URL for the repo's zipball."""
 
+        self.ref = check_default_branch(self.owner, self.repo, "gitlab")
+
         return "https://gitlab.com/{owner}/{repo}/-/archive/{ref}/{repo}-{ref}.zip".format(
             owner=self.owner, repo=self.repo, ref=self.ref
         )
@@ -1674,6 +1730,8 @@ class GitLabURL(RepoTypeURL):
     def compose_content_url(self, api=False, tree=False, default="{}"):
         """Compose GitLab URL for the content of a file or a directory."""
 
+        self.ref = check_default_branch(self.owner, self.repo, "gitlab")
+
         if api:
             if not tree:
                 return "https://gitlab.com/api/v4/projects/{}%2F{}/repository/files/{}/raw?ref={}".format(
@@ -1751,7 +1809,7 @@ class GitLabURL(RepoTypeURL):
                 self.repo,
                 self.ref,
                 self.path,
-            ) = RepoTypeURL.interpret_repo_ref(url)
+            ) = RepoTypeURL.interpret_repo_ref(url, "gitlab")
 
             if "#" in url:
                 self.ref = "merge_requests/" + self.ref + "/head"
@@ -1774,8 +1832,8 @@ class GitLabURL(RepoTypeURL):
                     self.ref = seg[2]
                 elif seg[0] == "merge_requests":
                     self.ref = (
-                        "/".join("/".join(seg[2:]).split("=")[1:])
-                        or "/".join(seg[:2]) + "/head"
+                            "/".join("/".join(seg[2:]).split("=")[1:])
+                            or "/".join(seg[:2]) + "/head"
                     )
 
         if self.path and self.path.endswith("/"):
@@ -1792,12 +1850,16 @@ class BitbucketURL(RepoTypeURL):
     def compose_repo_zip_url(self):
         """Compose Bitbucket URL for the repo's zipball."""
 
+        self.ref = check_default_branch(self.owner, self.repo, "bitbucket")
+
         return "https://bitbucket.org/{}/{}/get/{}.zip".format(
             self.owner, self.repo, self.ref
         )
 
     def compose_content_url(self, api=False, tree=False, default="{}"):
-        """Compose Bitbuckt URL for the content of a file or a directory."""
+        """Compose Bitbucket URL for the content of a file or a directory."""
+
+        self.ref = check_default_branch(self.owner, self.repo, "bitbucket")
 
         if api:
             return "https://api.bitbucket.org/2.0/repositories/{}/{}/src/{}/{}?format=meta".format(
@@ -1857,7 +1919,7 @@ class BitbucketURL(RepoTypeURL):
                 self.repo,
                 self.ref,
                 self.path,
-            ) = RepoTypeURL.interpret_repo_ref(url)
+            ) = RepoTypeURL.interpret_repo_ref(url, "bitbucket")
 
             if "#" in url:
                 self.ref = "pull-requests/" + self.ref + "/head"
@@ -2043,7 +2105,7 @@ config file path."""
 
 
 def gen_packages_yaml(
-    mlmodelsyaml="MLMODELS.yaml", packagesyaml="Packages.yaml"
+        mlmodelsyaml="MLMODELS.yaml", packagesyaml="Packages.yaml"
 ):
     """Generate Packages.yaml, the curated list of model packages, by just
 concatenate all MLHUB.yaml.  By default, it will generate
@@ -2103,7 +2165,7 @@ Packages.yaml in current working dir.
 
 
 def gen_packages_yaml2(
-    mlmodelsyaml="MLMODELS.yaml", packagesyaml="Packages.yaml"
+        mlmodelsyaml="MLMODELS.yaml", packagesyaml="Packages.yaml"
 ):
     """Generate Packages.yaml, the curated list of model packages, using
 yaml to ensure correct format.  By default, it will generate
@@ -2186,7 +2248,6 @@ def update_working_dir(model, folder):
 
 
 def update_sys_python_pkg_usage(model, usage=True):
-
     update_config(model, {SYS_PYTHON_PKG_USAGE: usage})
 
 
@@ -2220,7 +2281,6 @@ def get_sys_python_pkg_usage(model):
 
 
 def get_py_pkg_path_env(model):
-
     # 20200719 Version 20.1.1 of pip3 is installed on Ubuntu 20.04
     # after `pip3 install --upgrade pip`. This uses
     # site.getsitepackages()[0]
@@ -2232,14 +2292,14 @@ def get_py_pkg_path_env(model):
     # mlhub as the root location of the installed packages and so
     # different pip version cause confusion.
 
-    # 20200720 Tried to determin which to use here dynamically but
+    # 20200720 Tried to determine which to use here dynamically but
     # getting complicated.
 
     # 20200722 Now with `ml configure` resume upgrading to the latest
     # pip3. Then use python_pkg_base as the `--target .python`
     # location for pip3 to install the packages. Thus that is also on
     # PYTHONPATH. The bin is also .python/bin.
-    
+
     python_pkg_base = os.path.sep.join([get_package_dir(model), ".python"])
     python_pkg_bin = python_pkg_base + "/bin"
 
@@ -2252,16 +2312,17 @@ def get_py_pkg_path_env(model):
 
     # 20200722 Is this still needed? Under what scenario? Is this the
     # check for a conda environment?
-    
+
     if sys.executable != SYS_PYTHON_CMD:
         python_pkg_bin = python_pkg_base + site.PREFIXES[0] + "/bin"
         if get_sys_python_pkg_usage(model):
             print_on_stderr(MSG_INCOMPATIBLE_PYTHON_ENV, model)
 
-    exports  = f'export PATH="{python_pkg_bin}:$PATH"; '
+    exports = f'export PATH="{python_pkg_bin}:$PATH"; '
     exports += f'export PYTHONPATH="{python_pkg_base}"; '
-    
-    return(exports)
+
+    return (exports)
+
 
 # ----------------------------------------------------------------------
 # Bash completion helper
@@ -2351,9 +2412,9 @@ def get_model_completion_list():
 def find_best_match(misspelled, candidates):
     """Find the best matched word with <misspelled> in <candidates>."""
 
-    best_match = fuzzprocess.extract(
+    best_match = fuzzprocess.extractOne(
         misspelled, candidates, scorer=fuzz.ratio
-    )[0]
+    )
     matched = best_match[0]
     score = best_match[1]
 
@@ -2367,7 +2428,6 @@ def is_misspelled(score):
 
 
 def get_misspelled_command(command, available_commands):
-
     matched, score = find_best_match(command, available_commands)
     if is_misspelled(score):
         yes = yes_or_no(
@@ -2384,7 +2444,6 @@ def get_misspelled_command(command, available_commands):
 
 
 def get_misspelled_pkg(model):
-
     model_completion_list = get_model_completion_list()
     if len(model_completion_list) != 0:
         matched, score = find_best_match(model, model_completion_list)
@@ -2428,7 +2487,7 @@ corresponding functions defined in <module>."""
 
         cmd_meta = self.commands[subcommand]
         self.logger.debug(
-            "Add command line positioanl argument: {} - {}".format(
+            "Add command line positional argument: {} - {}".format(
                 subcommand, cmd_meta
             )
         )
